/* Copyright (C) 2006 Andre Noll <maan@tuebingen.mpg.de>, see file COPYING. */

/** \file dccp_recv.c paraslash's dccp receiver */

/*
 * based on client.c of dccp-cs-0.01.tar.bz2,
 * (C) 2005 Ian McDonald <imcdnzl@gmail.com>
 */

#include <netinet/in.h>
#include <sys/socket.h>
#include <regex.h>
#include <sys/types.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <netdb.h>
#include <lopsub.h>

#include "recv_cmd.lsg.h"
#include "para.h"
#include "error.h"
#include "list.h"
#include "sched.h"
#include "buffer_tree.h"
#include "recv.h"
#include "string.h"
#include "net.h"
#include "fd.h"

static void dccp_recv_close(struct receiver_node *rn)
{
	if (rn->fd > 0)
		close(rn->fd);
	btr_pool_free(rn->btrp);
}

/* Check whether the host supports the requested 'ccid' arguments. */
static int dccp_recv_ccid_support_check(const struct lls_parse_result *lpr)
{
	uint8_t *ccids;
	int i, j, ret, nccids;
	unsigned given = RECV_CMD_OPT_GIVEN(DCCP, CCID, lpr);

	ret = dccp_available_ccids(&ccids);
	if (ret < 0)
		return ret;
	nccids = ret;
	for (i = 0; i < given; i++) {
		uint32_t val = lls_uint32_val(i,
			RECV_CMD_OPT_RESULT(DCCP, CCID, lpr));
		for (j = 0; j < nccids && ccids[j] != val; j++)
			;
		if (j == nccids) {
			PARA_ERROR_LOG("'CCID-%u' not supported on this host\n",
				val);
			return -ERRNO_TO_PARA_ERROR(EINVAL);
		}
	}
	return 1;
}

static int dccp_recv_open(struct receiver_node *rn)
{
	struct lls_parse_result *lpr = rn->lpr;
	struct flowopts *fo = NULL;
	uint8_t *ccids = NULL;
	int fd, ret, i;
	const struct lls_opt_result *r_c = RECV_CMD_OPT_RESULT(DCCP, CCID, lpr);
	const char *host = RECV_CMD_OPT_STRING_VAL(DCCP, HOST, lpr);
	uint32_t port = RECV_CMD_OPT_UINT32_VAL(DCCP, PORT, lpr);
	unsigned given;

	ret = dccp_recv_ccid_support_check(lpr);
	if (ret < 0)
		return ret;
	/* Copy CCID preference list (u8 array required) */
	given = lls_opt_given(r_c);
	if (given) {
		ccids = para_malloc(given);
		fo = flowopt_new();
		for (i = 0; i < given; i++)
			ccids[i] = lls_int32_val(i, r_c);
		OPT_ADD(fo, SOL_DCCP, DCCP_SOCKOPT_CCID, ccids, i);
	}

	fd = makesock(IPPROTO_DCCP, 0, host, port, fo);
	flowopt_cleanup(fo);
	free(ccids);
	if (fd < 0)
		return fd;
	/*
	 * Disable unused CCIDs: the receiver does not send any application
	 * data to the server. By shutting down this unused path we reduce
	 * internal processing costs, as the unused CCIDs (in the kernel) are
	 * then bypassed.
	 */
	if (shutdown(fd, SHUT_WR) < 0) {
		ret = -ERRNO_TO_PARA_ERROR(errno);
		goto err;
	}
	ret = mark_fd_nonblocking(fd);
	if (ret < 0)
		goto err;
	rn->btrp = btr_pool_new("dccp_recv", 320 * 1024);
	rn->fd = fd;
	return 1;
err:
	close(fd);
	return ret;
}

static void dccp_recv_pre_select(struct sched *s, void *context)
{
	struct receiver_node *rn = context;

	if (generic_recv_pre_select(s, rn) <= 0)
		return;
	para_fd_set(rn->fd, &s->rfds, &s->max_fileno);
}

static int dccp_recv_post_select(struct sched *s, void *context)
{
	struct receiver_node *rn = context;
	struct btr_node *btrn = rn->btrn;
	struct iovec iov[2];
	int ret, iovcnt;
	size_t num_bytes;

	ret = task_get_notification(rn->task);
	if (ret < 0)
		goto out;
	ret = btr_node_status(btrn, 0, BTR_NT_ROOT);
	if (ret <= 0)
		goto out;
	iovcnt = btr_pool_get_buffers(rn->btrp, iov);
	ret = -E_DCCP_OVERRUN;
	if (iovcnt == 0)
		goto out;
	ret = readv_nonblock(rn->fd, iov, iovcnt, &s->rfds, &num_bytes);
	if (num_bytes == 0)
		goto out;
	if (num_bytes <= iov[0].iov_len) /* only the first buffer was filled */
		btr_add_output_pool(rn->btrp, num_bytes, btrn);
	else { /* both buffers contain data */
		btr_add_output_pool(rn->btrp, iov[0].iov_len, btrn);
		btr_add_output_pool(rn->btrp, num_bytes - iov[0].iov_len, btrn);
	}
out:
	if (ret < 0)
		btr_remove_node(&rn->btrn);
	return ret;
}

const struct receiver lsg_recv_cmd_com_dccp_user_data = {
	.open = dccp_recv_open,
	.close = dccp_recv_close,
	.pre_select = dccp_recv_pre_select,
	.post_select = dccp_recv_post_select,
};
