/* Copyright (C) 2006 Andre Noll <maan@tuebingen.mpg.de>, see file COPYING. */

/** \file write.h Writer-related structures. */

/**
 * Describes one running instance of a writer.
 */
struct writer_node {
	/** The ID of this writer. */
	int wid;
	/** Writer-specific data. */
	void *private_data;
	/** The parsed command line, merged with options given in the config file. */
	struct lls_parse_result *lpr;
	/** The buffer tree node associated with this writer node. */
	struct btr_node *btrn;
	/** The task of this writer node. */
	struct task *task;
	/** The minimal input queue size (size of one audio sample). */
	size_t min_iqs;
};

/** Describes one supported writer. */
struct writer {
	/**
	 * Prepare the fd sets for select.
	 *
	 * This is called from scheduler. It may use the sched pointer to add
	 * any file descriptors or to decrease the select timeout.
	 */
	void (*pre_select)(struct sched *s, void *context);
	/**
	 * Write audio data.
	 *
	 * Called from the post_select function of the writer node's task.
	 */
	int (*post_select)(struct sched *s, void *context);
	/**
	 * Close one instance of the writer.
	 *
	 * This function is assumed to succeed.
	 */
	void (*close)(struct writer_node *);
	/**
	 * The callback handler.
	 *
	 * Each writer may provide an ->execute callback which can be used for
	 * inter-node communication.
	 */
	btr_command_handler execute;
};

#define WRITE_CMD(_num) (lls_cmd(_num, write_cmd_suite))

#define WRITE_CMD_OPT_RESULT(_cmd, _opt, _lpr) \
	(lls_opt_result(LSG_WRITE_CMD_ ## _cmd ## _OPT_ ## _opt, _lpr))
#define WRITE_CMD_OPT_GIVEN(_cmd, _opt, _lpr) \
	(lls_opt_given(WRITE_CMD_OPT_RESULT(_cmd, _opt, _lpr)))
#define WRITE_CMD_OPT_UINT32_VAL(_cmd, _opt, _lpr) \
	(lls_uint32_val(0, WRITE_CMD_OPT_RESULT(_cmd, _opt, (_lpr))))
#define WRITE_CMD_OPT_STRING_VAL(_cmd, _opt, _lpr) \
	(lls_string_val(0, WRITE_CMD_OPT_RESULT(_cmd, _opt, (_lpr))))

int check_writer_arg_or_die(const char *wa, struct lls_parse_result **lprp);
const struct writer *writer_get(int wid);
const char *writer_name(int wid);
void register_writer_node(struct writer_node *wn, struct btr_node *parent,
		struct sched *s);
int get_btr_sample_rate(struct btr_node *btrn, int32_t *result);
int get_btr_channels(struct btr_node *btrn, int32_t *result);
int get_btr_sample_format(struct btr_node *btrn, int32_t *result);
void print_writer_helps(bool detailed);
